# TBD: refactoring as class (i.e. Trainer()) necessary?
# TBD: explicit scikit learn pipeline required?
#   --> Keras workflow is already a pipeline
# TODO: model.predict function in separate file

import joblib
import pandas as pd
from projectYoda.data import get_dataframes
from projectYoda.gcp import store_model_on_gcp
from projectYoda.params import root_dir
from tensorflow.keras.preprocessing.image import ImageDataGenerator
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Flatten, Conv2D, Dropout, MaxPooling2D, Activation
from tensorflow.keras import optimizers
from tensorflow.keras import callbacks
from termcolor import colored


# Source for image data preprocessing
# https://vijayabhaskar96.medium.com/tutorial-on-keras-imagedatagenerator-with-flow-from-dataframe-8bd5776e45c1

def preprocess_data(df_train, df_valid, root_dir):
    '''Loads and preprocesses image data'''

    train_datagen = ImageDataGenerator(rescale = 1./255.,
                                       rotation_range = 30,
                                       width_shift_range = 0.3,
                                       height_shift_range = 0.3,
                                       brightness_range=[0.2,1.0],
                                       shear_range = 0.3,
                                       zoom_range = 0.4,
                                       fill_mode='nearest',
                                       vertical_flip=True,
                                       horizontal_flip = True)
    test_datagen = ImageDataGenerator(rescale=1.0 / 255.)

    train_generator = train_datagen.flow_from_dataframe(dataframe=df_train,
                                                directory=root_dir,
                                                x_col="path",
                                                y_col="minifigure_name",
                                                class_mode="categorical",
                                                target_size=(512, 512),
                                                batch_size=16)
    valid_generator = test_datagen.flow_from_dataframe(dataframe=df_valid,
                                                directory=root_dir,
                                                x_col="path",
                                                y_col="minifigure_name",
                                                class_mode="categorical",
                                                target_size=(512, 512),
                                                batch_size=16,
                                                shuffle=False)
    return train_generator, valid_generator


# TODO: finetune model params and try different models with transfer learning
def init_model():
    '''Initializes model'''

    # params - to be finetuned
    input_shape = (512, 512, 3)
    padding = 'same'
    number_of_classes = 36
    opt = optimizers.Adam(learning_rate=0.001)

    # archtitecture
    model = Sequential()
    model.add(Conv2D(32, (3, 3), padding=padding, input_shape=input_shape))
    model.add(Activation('relu'))
    model.add(Conv2D(32, (3, 3)))
    model.add(Activation('relu'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Dropout(0.25))

    model.add(Conv2D(64, (3, 3), padding=padding))
    model.add(Activation('relu'))
    model.add(Conv2D(64, (3, 3)))
    model.add(Activation('relu'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Dropout(0.25))

    model.add(Flatten())
    model.add(Dense(512))
    model.add(Activation('relu'))
    model.add(Dropout(0.5))
    model.add(Dense(number_of_classes, activation='softmax'))

    model.compile(optimizer=opt,
                loss="categorical_crossentropy",
                metrics=["accuracy"])
    return model


# TODO: finetune hyperparams
def train(model, train_generator, valid_generator):
    '''Trains model on '''

    epochs = 50
    es = callbacks.EarlyStopping(patience=2,
                                 monitor='val_accuracy',
                                 restore_best_weights=True)


    model.fit_generator(generator=train_generator,
                    validation_data=valid_generator,
                    callbacks=[es],
                    verbose=1,
                    epochs=epochs)
    # save model locally
    joblib.dump(model, 'model.joblib')
    print(colored("model.joblib saved locally", "green"))
    store_model_on_gcp()
    return model

# TODO: implement function to evaluate model on test set
'''def evaluate(model, test_generator):
        """evaluates the pipeline on df_test and return the RMSE"""
        scores = model.score(test_generator)
        accuracy = # TBD
        return accuracy'''


if __name__ == "__main__":
    df_train, df_valid = get_dataframes()
    train_generator, valid_generator = preprocess_data(df_train, df_valid, root_dir)
    model = init_model()
    trained_model = train(model, train_generator, valid_generator)



import os
from google.cloud import storage
from termcolor import colored
from projectYoda.params import BUCKET_NAME, MODEL_NAME, MODEL_VERSION, PATH_TO_LOCAL_MODEL
import joblib


LOCAL_MODEL_NAME = PATH_TO_LOCAL_MODEL

def store_model_on_gcp(rm=False):
    '''Upload model to Google Cloud Storage'''
    client = storage.Client().bucket(BUCKET_NAME)

    storage_location = f"models/{MODEL_NAME}/{MODEL_VERSION}/{LOCAL_MODEL_NAME}"
    blob = client.blob(storage_location)
    blob.upload_from_filename('model.joblib')
    print(colored(f"=> model.joblib uploaded to bucket {BUCKET_NAME} inside {storage_location}",
                  "green"))
    if rm:
        os.remove('model.joblib')


def get_model_from_gcp(rm=False):
    """Get the model from Google Cloud Storage"""
    storage_location = f"models/{MODEL_NAME}/{MODEL_VERSION}/{LOCAL_MODEL_NAME}"
    path = f"gs://{BUCKET_NAME}/{storage_location}"
    model = joblib.load(path)
    return model



from google.cloud import storage
import pandas as pd
from PIL import Image as PImage
from projectYoda.params import root_dir, BUCKET_NAME, BUCKET_DATA_PATH

def get_dataframes_locally():
    '''Get train and validation dataframes'''
    index = pd.read_csv('../raw_data/index.csv')
    metadata = pd.read_csv('../raw_data/metadata.csv')
    df = pd.merge(index, metadata, on='class_id')

    # put all but 2 images of each class in training set
    df_train = [
        df[df['class_id'] == i].iloc[:-2]
        for i in range(1,
                       len(df['class_id'].value_counts()) + 1)
    ]

    # put 2 images of each class in validation set
    df_valid = [
        df[df['class_id'] == i].iloc[-2:]
        for i in range(1,
                       len(df['class_id'].value_counts()) + 1)
    ]

    return df_train, df_valid


def get_dataframes_from_gcp(optimize=False, **kwargs):
    """method to get the training data (or a portion of it) from google cloud bucket"""
    client = storage.Client()
    base_path = f"gs://{BUCKET_NAME}/{BUCKET_DATA_PATH}"
    index = pd.read_csv(f'{base_path}/index.csv')
    metadata = pd.read_csv(f'{base_path}/metadata.csv')
    df = pd.merge(index, metadata, on='class_id')

    # put all but 2 images of each class in training set
    df_train = [
        df[df['class_id'] == i].iloc[:-2]
        for i in range(1,
                       len(df['class_id'].value_counts()) + 1)
    ]

    # put 2 images of each class in validation set
    df_valid = [
        df[df['class_id'] == i].iloc[-2:]
        for i in range(1,
                       len(df['class_id'].value_counts()) + 1)
    ]

    return df_train, df_valid


def load_images(root_dir, df):
    '''Load images from root director and path from datafram'''

    loaded_images = [PImage.open(root_dir + image) for image in df['path']]

    return loaded_images

if __name__ == "__main__":
    df_train, df_valid = get_dataframes_from_gcp()
    print(type(df_valid))



### MLFLOW configuration - - - - - - - - - - - - - - - - - - -

MLFLOW_URI = "https://mlflow.lewagon.co/"
EXPERIMENT_NAME = "[SW] [Stockholm] [steiningerjakob] prod_v0.1"

### DATA & MODEL LOCATIONS  - - - - - - - - - - - - - - - - - - -

PATH_TO_LOCAL_MODEL = 'model.joblib'

root_dir = "../raw_data/"

### GCP configuration - - - - - - - - - - - - - - - - - - -

# /!\ you should fill these according to your account

### GCP Project - - - - - - - - - - - - - - - - - - - - - -

# not required here

### GCP Storage - - - - - - - - - - - - - - - - - - - - - -

BUCKET_NAME = 'wagon-data-745-project-yoda'

##### Data  - - - - - - - - - - - - - - - - - - - - - - - -

# train data file location - not yet implemented
BUCKET_DATA_PATH = 'raw_data'

##### Training  - - - - - - - - - - - - - - - - - - - - - -

# not required here

##### Model - - - - - - - - - - - - - - - - - - - - - - - -

# model folder name (will contain the folders for all trained model versions)
MODEL_NAME = 'classification'

# model version folder name (where the trained model.joblib file will be stored)
MODEL_VERSION = 'v1'

### GCP AI Platform - - - - - - - - - - - - - - - - - - - -

# not required here

### - - - - - - - - - - - - - - - - - - - - - - - - - - - -



from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

app = FastAPI()

app.add_middleware(
    CORSMiddleware,
    allow_origins=['*'],
    allow_credentials=True,
    allow_methods=['*'],
    allow_headers=['*'],
)

@app.get('/')
def index():
    return {'Yoda says:', 'You must unlearn what you have learned!'}



# PSEUDO CODE
# load the model.joblib file with basic model
# use pipeline to preprocess the data
# classification = model.predict(<picutre vectorized>)
# result = classifciation[0]
# return dict(prediction = pred)
